import logging
from time import perf_counter

import networkx as nx

from evrouting.T import Result
from evrouting import gasstation, charge
from evrouting.graph_tools import (
    consumption_function_distance_factory,
    DISTANCE_KEY
)
from evrouting.osm.profiles import car
from evrouting.osm.routing import GasstationAccessFunctions, a_start_heuristic

from lib.T import (
    GasstationQueryRow,
    ChargeQueryRow,
    ClassicQueryRow,
    AStarQueryRow
)

__all__ = [
    'gasstation_query',
    'charge_query',
    'classic_query',
    'astar_query'
]

CACHE = {}


def cached(key):
    """Cache return value."""
    def decorator(func):
        def getter(*args, **kwargs):
            try:
                return CACHE[key]
            except KeyError:
                ret = func(*args, **kwargs)
                CACHE[key] = ret
                return ret

        return getter

    return decorator


@cached('gasstation')
def get_contracted_graph(graph, conf, f):
    start = perf_counter()
    contracted_graph = gasstation.routing.contract_graph(
        G=graph,
        charging_stations=graph.charging_stations,
        capacity=conf['capacity'],
        f=f
    )
    contraction_time = perf_counter() - start
    logging.info('Contracted graph with {} stations in {:.2f} s.'.format(
        len(graph.charging_stations),
        contraction_time
    ))
    return contracted_graph, contraction_time


@cached('state_graph')
def get_state_graph(contracted_graph, conf, f):
    start = perf_counter()
    state_graph = gasstation.routing.state_graph(
        contracted_graph,
        conf['capacity'],
        f
    )
    state_graph_time = perf_counter() - start
    logging.info('Created State Graph in {:.2f} s.'.format(
        state_graph_time
    ))
    return state_graph, state_graph_time


def gasstation_query(graph, conf, s, t):
    f = GasstationAccessFunctions(conf['consumption']['consumption_coefficient'])
    contracted_graph, contraction_time = get_contracted_graph(graph, conf, f)
    state_graph, state_graph_time = get_state_graph(contracted_graph, conf, f)

    start = perf_counter()
    result = gasstation.routing.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        f=f,
        extended_graph=state_graph,
        contracted_graph=contracted_graph
    )
    query_time = perf_counter() - start

    return GasstationQueryRow(
        start_node=s,
        target_node=t,
        query_time=query_time,
        trip_time=result.trip_time if type(result) == Result else None,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        time_contracted_graph=contraction_time,
        time_state_graph=state_graph_time
    )


def charge_query(graph, conf, s, t):
    start = perf_counter()
    result = charge.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        c=consumption_function_distance_factory(conf['consumption']['consumption_coefficient'])
    )
    runtime = perf_counter() - start

    return ChargeQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result.trip_time if type(result) == Result else None,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations)
    )


def classic_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result = nx.shortest_path_length(graph, s, t, weight=DISTANCE_KEY)
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return ClassicQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        dijkstra_rank=0
    )


def astar_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result = nx.astar_path_length(
            graph,
            s, t,
            weight=DISTANCE_KEY,
            heuristic=a_start_heuristic(graph, car)
        )
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return AStarQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations)
    )
